/*	      Directory class for reading directory entries on linux
 *	      Copyright (C) 2003, Piotr Kalinowski <pitkali@tlen.pl>
 *	      All rights reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <sys/types.h>
#include <dirent.h>
#include <cstring>
#include <cstdlib>
#include <errno.h>
#include <limits>
#include "directory.hpp"

// -------------------------- _directory ----------------------------

// This class is the real system-dependent implementation of Directory
class _directory {
  public:
    _directory():_dir(NULL),_ref(1) {}
    _directory(const std::string& s):_dir(NULL),_ref(1) { open(s); }
    ~_directory() { close(); }
    
    bool operator==(const _directory& rhs) { return (_dir == rhs._dir); }
    
    bool open(const std::string& s) // open directory
    { 
      if (_dir)
	close();
      if ((_dir = opendir(s.c_str())) == NULL) 
	_error = strerror(errno);
      else
	_name = s;
      return isopen();
    }

    void close(void) 
    { 
      if (isopen()) { 
	closedir(_dir); 
	_dir = NULL;
      } 
      _error.clear(); 
      _name.clear();
    }
    
    bool isopen(void) const { return (_dir != NULL); }
    bool isbad(void) const { return (_dir == NULL); }
    const std::string& get_error(void) const { return _error; }
    const std::string& get_name(void) const { return _name; }

    off_t tell(void) const { return telldir(_dir); } // returns -1 on error
    void seek(off_t off) const { if (isopen()) seekdir(_dir, off); }
    void rewind(void) const { if (isopen()) rewinddir(_dir); }
    
    std::string read_entry(void) const
    {
      dirent* entry = readdir(_dir);
      if (entry)
	return std::string(entry->d_name);
      else
	return std::string();
    }

    unsigned short ref(void) const { return _ref; }
    void inc_ref(void) const { ++_ref; }
    void dec_ref(void) const { --_ref; }

  private:
    // copying and assigning is forbidden
    _directory(const _directory& rhs);
    _directory& operator=(const _directory& rhs);

    DIR* _dir; // underlying dir structure
    std::string _error; // error info
    std::string _name; // opened dir name
    mutable unsigned long _ref;
};

// --------------------------------- Directory ----------------------

// the interface comes here

Directory::Directory()
{ impl = new _directory; }

Directory::Directory(const std::string& s)
{ impl = new _directory(s); }

Directory::~Directory()
{ 
  impl->dec_ref();
  if (impl->ref() == 0)
    delete impl;
}

Directory::Directory(const Directory& rhs)
{
  impl = rhs.impl;
  impl->inc_ref();
}

Directory& Directory::operator=(const Directory& rhs)
{
  impl->dec_ref();
  if (impl->ref() == 0)
    delete impl;
  impl = rhs.impl;
  impl->inc_ref();
  return *this;
}

bool Directory::operator==(const Directory& rhs) { return *impl == *(rhs.impl); }

bool Directory::open(const std::string& s) 
{ 
  if (impl->isopen())
    close();
  return impl->open(s); 
}

void Directory::close(void) 
{ 
  if (impl->ref() == 1)
    impl->close();
  else {
    impl->dec_ref();
    impl = new _directory;
  }
}

bool Directory::isopen(void) const { return impl->isopen(); }
bool Directory::isbad(void) const { return impl->isbad(); }

std::string Directory::read_entry(void) const { return impl->read_entry(); }
off_t Directory::tell(void) const { return impl->tell(); }
void Directory::seek(off_t off) const { impl->seek(off); }
void Directory::rewind(void) const { impl->rewind(); }

const std::string& Directory::get_error(void) const { return impl->get_error(); }
const std::string& Directory::get_name(void) const { return impl->get_name(); }

