/*
 *		RegExp - class encapsulating GNU regular expressions
 *		Copyright (C) 2003, Piotr Kalinowski <pitkali@tlen.pl>
 *		All rights reserved.
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit
 * persons to whom the Software is furnished to do so, subject to the
 * following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 */

#include <limits>
#include <new>
#include <algorithm>
#include <cstdlib>
#include <cctype>
#include "regexp.hpp"

using namespace std;

const unsigned int nchars = numeric_limits<unsigned char>::max()+1;

// ---------------------------- Match -------------------------------

// Copy constructor for Match
RegExp::Match::Match(const RegExp::Match& rhs) throw(bad_alloc)
{
  if (rhs._regs.num_regs) { // rhs is not empty
    _regs.num_regs = rhs._regs.num_regs;
    // using malloc for consistency whith regex functions
    _regs.start = static_cast<regoff_t*> (malloc(_regs.num_regs*sizeof(regoff_t)));
    if (_regs.start == NULL) {
      _regs.start = _regs.end = 0;
      _regs.num_regs = 0;
      throw bad_alloc();
    }
    _regs.end = static_cast<regoff_t*> (malloc(_regs.num_regs*sizeof(regoff_t)));
    if (_regs.end == NULL) {
      free(_regs.start);
      _regs.start = _regs.end = 0;
      _regs.num_regs = 0;
      throw bad_alloc();
    }
    for (register unsigned int i = 0; i<_regs.num_regs; ++i) {
      _regs.start[i] = rhs._regs.start[i];
      _regs.end[i] = rhs._regs.end[i];
    }
  }
  else {
    _regs.num_regs = 0;
    _regs.start = _regs.end = 0;
  }
}

RegExp::Match& RegExp::Match::operator=(const RegExp::Match& rhs) throw(bad_alloc)
{
  if (_regs.num_regs) {
    // delete what we already keep
    free(_regs.start);
    free(_regs.end);
  }
  if (rhs._regs.num_regs) { // rhs is not empty
    _regs.num_regs = rhs._regs.num_regs;
    // using malloc for consistency whith regex functions
    _regs.start = static_cast<regoff_t*> (malloc(_regs.num_regs*sizeof(regoff_t)));
    if (_regs.start == NULL) {
      _regs.start = _regs.end = 0;
      _regs.num_regs = 0;
      throw bad_alloc();
    }
    _regs.end = static_cast<regoff_t*> (malloc(_regs.num_regs*sizeof(regoff_t)));
    if (_regs.end == NULL) {
      free(_regs.start);
      _regs.start = _regs.end = 0;
      _regs.num_regs = 0;
      throw bad_alloc();
    }
    for (register unsigned int i = 0; i<_regs.num_regs; ++i) {
      _regs.start[i] = rhs._regs.start[i];
      _regs.end[i] = rhs._regs.end[i];
    }
  }
  else {
    _regs.num_regs = 0;
    _regs.start = _regs.end = 0;
  }
  return *this;
}

// -------------------------- RegExp --------------------------------

// turn using fastmap on or off
bool RegExp::use_fastmap(bool f)
{ 
  if (!f) {
    if (_buf.fastmap) {
      // deallocate fastmap
      free(_buf.fastmap);
      _buf.fastmap = 0;
    }
    return true;
  }
  if (_buf.fastmap) // we already have fastmap
    return true;
  // allocate memory for fastmap
  _buf.fastmap = static_cast<char*>(malloc(sizeof(char)*nchars));
  if (_buf.fastmap == NULL) { // whoops - no memory
    _buf.fastmap = 0; 
    return false; 
  }
  return true;
}

//turn translate table on or off
bool RegExp::use_translate(const std::string& t) throw(invalid_argument)
{
  if (t.size() == 0) { // turn translate off
    if (_buf.translate) { // deallocate memory
      free(_buf.translate);
      _buf.translate = 0;
    }
    return true; // success
  }
  if (t.size() != nchars) {
    throw invalid_argument("Wrong number of characters in translate table!");
  }
  if (_buf.translate) // delete previous table
    free(_buf.translate);
  _buf.translate = static_cast<char*>(malloc(sizeof(char)*t.size()));  // alocate mem
  if(_buf.translate == NULL) { // whooops - no mem
    _buf.translate = 0;
    return false;
  }
  for (register unsigned int i = 0; i<t.size(); ++i)
    // copy given translate table
    _buf.translate[i] = t[i];
  return true;
}

bool RegExp::ignore_case(void)
{
  string ic(nchars, ' ');
  for (register unsigned int i = 0; i<nchars; ++i)
    ic[i] = tolower(i);
  return use_translate(ic);
}

// compile with not default syntax
bool RegExp::compile(const string& s, reg_syntax_t syn)
{
  reg_syntax_t old = syntax();
  syntax() = syn;
  bool retval = compile(s);
  syntax() = old;
  return retval;
}

// compile regular expression
bool RegExp::compile(const string& s)
{
  if (is_compiled()) {
    free(_buf.buffer);
    _buf.buffer = 0;
    _buf.allocated = 0;
    _m.clear();
  }
  const char* error = re_compile_pattern(s.c_str(),s.size(),&_buf);
  if (error == 0)
    return true;
  _err = error;
  _buf.buffer = 0;
  return false;
}

// *** Match expression ***
int RegExp::match(const string& s,const int start) 
throw(logic_error,runtime_error)
{
  if (!is_compiled())
    throw logic_error("No expression has been compiled yet.");
  int retval =  re_match(&_buf,s.c_str(),s.size(), start,&(_m._regs));
  if (retval == -2)
    throw runtime_error("re_match internal error occurred.");
  return retval;
}
// *** Search expression ***
int RegExp::search(const string& s,const int start,const int range) 
throw(logic_error,runtime_error)
{
  if (!is_compiled())
    throw logic_error("No expression has been compiled yet.");
  int retval =  re_search(&_buf,s.c_str(),s.size(), start,
      (range == 0)? (s.size()) : range,&(_m._regs));
  if (retval == -2)
    throw runtime_error("re_search internal error occurred.");
  return retval;
}
